use strict;
use warnings;

# this test was generated with Dist::Zilla::Plugin::Test::EOL 0.18

use Test::More 0.88;
use Test::EOL;

my @files = (
    'Build.PL',
    'COPYING',
    'Changes',
    'GPLv3',
    'META.json',
    'META.yml',
    'README',
    'lib/Dist/Zilla/Role/TextTemplater.pm',
    'lib/Dist/Zilla/Role/TextTemplater/Manual.pod',
    'lib/Dist/Zilla/Role/TextTemplater/ReadMe.pod',
    't/00-compile.t',
    't/DistZillaTester.pm',
    't/PluginTester.pm',
    't/TextTemplaterTestPlugin.pm',
    't/TextTemplaterTester.pm',
    't/text-templater.t',
    'xt/aspell.en.pws',
    'xt/author/critic.t',
    'xt/author/eol.t',
    'xt/author/no-tabs.t',
    'xt/author/pod-spell.t',
    'xt/author/test-version.t',
    'xt/perlcritic.ini',
    'xt/release/cpan-changes.t',
    'xt/release/dist-manifest.t',
    'xt/release/fixme.t',
    'xt/release/kwalitee.t',
    'xt/release/meta-json.t',
    'xt/release/minimum-version.t',
    'xt/release/mojibake.t',
    'xt/release/new-version.t',
    'xt/release/pod-coverage.t',
    'xt/release/pod-no404s.t',
    'xt/release/pod-syntax.t'
);

eol_unix_ok($_, { trailing_whitespace => 1 }) foreach @files;
done_testing;
